from typing import List
from abc import ABC, abstractmethod
from warnings import warn

import numpy as np

from .typeutils import (RandomStateType, check_random_state,
                        NotEnoughSamplesWarning)


class BaseQuerySampler(ABC):
    """Abstract Base Class for query samplers
    
    A query sampler is an object that takes as input labeled and/or unlabeled
    samples and use knowledge from them to selected the most informative ones.

    Args:
        batch_size: Numbers of samples to select.
    """
    def __init__(self, batch_size: int):
        self.batch_size = batch_size

    @abstractmethod
    def fit(self, X: np.ndarray, y: np.ndarray = None):
        """Fit the model on labeled samples.

        Args:
            X: Labeled samples of shape (n_samples, n_features).
            y: Labels of shape (n_samples).
        
        Returns:
            The object itself
        """
        pass

    @abstractmethod
    def select_samples(self, X: np.array) -> np.array:
        """Selects the samples to annotate from unlabeled data.

        Args:
            X: Pool of unlabeled samples of shape (n_samples, n_features).

        Returns:
            Indices of the selected samples of shape (batch_size).
        """
        pass

    def _not_enough_samples(self, X: np.array) -> bool:
        cond = X.shape[0] < self.batch_size
        if cond:
            warn(f'''Requested {self.batch_size} samples but data only
             has {X.shape[0]}. All available data will be returned''',
                 NotEnoughSamplesWarning)
        return cond


class ScoredQuerySampler(BaseQuerySampler):
    """Abstract Base Class handling query samplers relying on a total order.
    Query sampling methods often scores all the samples and then pick samples
    using these scores. This base class handles the selection system, only
    a scoring method is then required.

    Args:
        batch_size: Numbers of samples to select.
        strategy: Describes how to select the samples based on scores. Can be
                  "top", "weighted".
        random_state: Random seeding
    """
    def __init__(self, batch_size: int, strategy: str = 'top',
                 random_state: RandomStateType = None):
        super().__init__(batch_size)
        self.strategy = strategy
        self.random_state = check_random_state(random_state)

    @abstractmethod
    def score_samples(self, X: np.array) -> np.array:
        """Give an informativeness score to unlabeled samples.

        Args:
            X: Samples to evaluate.

        Returns:
            Scores of the samples.
        """
        pass

    def select_samples(self, X: np.array) -> np.array:
        """Selects the samples from unlabeled data using the internal scoring.

        Args:
            X: Pool of unlabeled samples of shape (n_samples, n_features).
            strategy: Strategy to use to select queries. Can be one of top,
                      linear_choice, or squared_choice.

        Returns:
            Indices of the selected samples of shape (batch_size).
        """
        if self._not_enough_samples(X):
            return np.arange(X.shape[0])

        sample_scores = self.score_samples(X)
        self.sample_scores_ = sample_scores
        if self.strategy == 'top':
            index = np.argsort(sample_scores)[-self.batch_size:]
        elif self.strategy == 'weighted':
            index = self.random_state.choice(
                np.arange(X.shape[0]), size=self.batch_size,
                replace=False, p=sample_scores / np.sum(sample_scores))
        else:
            raise ValueError('Unknown sample selection strategy {}'
                             .format(self.strategy))
        return index



# Note: This code is inspired from modAL implementation
# https://modal-python.readthedocs.io/en/latest/content/query_strategies/ranked_batch_mode.html

import numpy as np

from .version import check_modules
check_modules('sklearn', 'batch')  # noqa

from sklearn.metrics import pairwise_distances, pairwise_distances_argmin_min

from .base import BaseQuerySampler


class RankedBatchSampler(BaseQuerySampler):
    """Selects samples to label by maximizing the distance between them.

    Args:
        batch_size: Number of samples to select.
        metric: Metric to use for distance computation.
    """
    def __init__(self, batch_size: int, metric: str = 'euclidean'):
        super().__init__(batch_size)
        self.metric = metric

    def fit(self, X: np.array, y: np.array = None) -> 'RankedBatchSampler':
        """Does nothing, RankedBatch is unsupervised.

        Args:
            X: Labeled samples of shape (n_samples, n_features).
            y: Labels of shape (n_samples).
       
        Returns:
            The object itself
        """
        return self

    def select_samples(self, X: np.array,
                       samples_weights: np.array) -> np.array:
        """Selects the samples to annotate from unlabelled data.
        
        Args:
            X: Pool of unlabeled samples of shape (n_samples, n_features).
            sample_weights: Weights of the samples of shape (n_samples).
                Set labeled samples as -1.

        Returns:
            Indices of the selected samples of shape (batch_size).
        """
        if self._not_enough_samples(X):
            return np.arange(X.shape[0])

        n_samples = X.shape[0]
        index = np.arange(n_samples)
        unlabeled_mask = (samples_weights > -.5)
        n_unlabeled = unlabeled_mask.sum()

        # We are going to modify this array so we copy it
        samples_weights = samples_weights.copy()

        # We compute the distances for labeled data in 2 steps
        _, similarity_scores = pairwise_distances_argmin_min(
            X[unlabeled_mask], X[np.logical_not(unlabeled_mask)],
            metric=self.metric)
        similarity_scores = 1 / (1 + similarity_scores)

        selected_samples = []

        for _ in range(self.batch_size):

            alpha = n_unlabeled / n_samples
            scores = (alpha * (1 - similarity_scores)
                      + (1 - alpha) * samples_weights[unlabeled_mask])

            idx_furthest = index[unlabeled_mask][np.argmax(scores)]
            selected_samples.append(idx_furthest)

            # Update similarities considering the selected sample as labeled
            # We could remove its value from the array but we avoid realloc
            sim = 1 / (1 + pairwise_distances(
                X[unlabeled_mask], X[idx_furthest, None],
                metric=self.metric)[:, 0])
            similarity_scores = np.max([similarity_scores, sim], axis=0)
            samples_weights[idx_furthest] = 0.
            n_unlabeled -= 1

        return np.asarray(selected_samples)



import os
from cardinal import cache
from tempfile import NamedTemporaryFile


def _test_backend(backend_class, suffix=''):
    f = NamedTemporaryFile()
    filename = f.name + suffix
    f.close()
    be = backend_class(filename)
    assert(len(be.get('test')) == 0)
    be._store('test', 37, key='example')
    be.close()
    be = backend_class(filename)
    assert be.get('test').to_dict() == {'key': {0: 'example'}, 'value': {0: 37}}
    be.close()
    os.unlink(filename)


def test_backends():
    _test_backend(cache.ShelveStore, '.db')
    _test_backend(cache.SqliteStore)



from copy import copy

from .version import check_modules

check_modules('examples', 'plotting')  # noqa

from matplotlib import pyplot as plt
from scipy.interpolate import make_interp_spline, BSpline
import numpy as np


def plot_confidence_interval(*args, label=None, q_inf=0.1, q_sup=0.9, alpha=.3, smoothing=None, dots=False):
    y_data = np.asarray(args[-1])
    if len(args) == 1:
        # We only have the data. We create x-axis values.
        x_data = np.arange(y_data.shape[0])
    else:
        x_data = np.asarray(args[0])

    avg = np.mean(y_data, axis=0)
    q10 = np.quantile(y_data, q_inf, axis=0)
    q90 = np.quantile(y_data, q_sup, axis=0)

    if smoothing is not None:
        x_plot = np.linspace(x_data.min(), x_data.max(), x_data.shape[0] * smoothing) 
        avg_plot = make_interp_spline(x_data, avg, k=2)(x_plot)
        q10_plot = make_interp_spline(x_data, q10, k=2)(x_plot)
        q90_plot = make_interp_spline(x_data, q90, k=2)(x_plot)
    else:
        x_plot = x_data
        avg_plot = avg
        q10_plot = q10
        q90_plot = q90
        
    line = plt.plot(x_plot, avg_plot, label=label)
    color = line[0].get_c()

    if dots:
        plt.scatter(x_data, avg, c=color)

    plt.fill_between(x_plot, q90_plot, q10_plot, color=color, alpha=alpha)



import numpy as np
from scipy.optimize import linear_sum_assignment

from .base import BaseQuerySampler
from .version import check_modules


class KCentroidSampler(BaseQuerySampler):
    """ KCentroid based query sampler.
    In order to increase diversity, it is possible to use a centroid based
    clustering to select samples.

    Args:
        clustering: A clustering algorithm matching the sklearn interface
        batch_size: Number of samples to draw when predicting.

    Attributes:
        clustering_ : The fitted clustering estimator.
    """
    def __init__(self, clustering, batch_size):
        super().__init__(batch_size)
        self.clustering_ = clustering

    def fit(self, X, y=None) -> 'KCentroidSampler':
        """Does nothing, this method is unsupervised.
        
        Args:
            X: Labeled samples of shape (n_samples, n_features).
            y: Labels of shape (n_samples).
        
        Returns:
            The object itself
        """
        return self

    def select_samples(self, X: np.array,
                       sample_weight: np.array = None) -> np.array:
        """Clusters the samples and select the ones closest to centroids.
        
        Args:
            X: Pool of unlabeled samples of shape (n_samples, n_features).
            sample_weight: Weight of the samples of shape (n_samples),
                optional.

        Returns:
            Indices of the selected samples of shape (batch_size).
        """
        if self._not_enough_samples(X):
            return np.arange(X.shape[0])

        kwargs = dict(sample_weight=sample_weight) if (sample_weight is not None) else dict()
        model = self.clustering_.fit(X, **kwargs)
        distances = model.transform(X)

        # Sometimes, one sample can be the closest to two centroids. In that
        # case, we want to take the second closest one, etc.
        # linear_sum_assignemnt solves this problem.
        return linear_sum_assignment(distances)[0]


class KMeansSampler(KCentroidSampler):
    """Select samples as closest sample to KMeans centroids.

    Args:
        batch_size: Number of samples to draw when predicting.
    """
    def __init__(self, batch_size, **kmeans_args):
        check_modules('sklearn', 'clustering.KmeansSampler')
        from sklearn.cluster import KMeans

        if 'n_clusters' in kmeans_args:
            raise ValueError(
                'You have specified n_clusters={} when creating KMeansSampler.'
                ' This is not supported since n_clusters is overridden using '
                'batch_size.'.format(kmeans_args['n_clusters']))
        kmeans_args['n_clusters'] = batch_size
        super().__init__(KMeans(**kmeans_args), batch_size)


class MiniBatchKMeansSampler(KCentroidSampler):
    """Select samples as closest sample to MiniBatchKMeans centroids.

    Args:
        batch_size: Number of samples to draw when predicting.
    """
    def __init__(self, batch_size, **kmeans_args):
        check_modules('sklearn', 'clustering.MiniBatchKmeansSampler')
        from sklearn.cluster import MiniBatchKMeans

        if 'n_clusters' in kmeans_args:
            raise ValueError(
                'You have specified n_clusters={} when creating '
                'MiniBatchKMeansSampler. This is not supported since'
                'n_clusters is overridden using '
                'batch_size.'.format(kmeans_args['n_clusters']))
        kmeans_args['n_clusters'] = batch_size
        super().__init__(MiniBatchKMeans(**kmeans_args), batch_size)



import numpy as np

from .version import check_modules

check_modules('sklearn', 'zhdanov2019')  # noqa

from .base import BaseQuerySampler
from .uncertainty import MarginSampler
from .clustering import KMeansSampler


class TwoStepKMeansSampler(BaseQuerySampler):
    """KMeans sampler using a margin uncertainty sampler as preselector

    """

    def __init__(self, beta: int, classifier, batch_size: int,
                 assume_fitted: bool = False, verbose: int = 0, **kmeans_args):
        
        self.sampler_list = [
            MarginSampler(classifier, beta * batch_size, strategy='top',
                          assume_fitted=assume_fitted, verbose=verbose),
            KMeansSampler(batch_size, **kmeans_args)
        ]

    def fit(self, X: np.array, y: np.array = None) -> 'TwoStepKMeansSampler':
        """Fits the first query sampler

        Args:
            X: Labeled samples of shape [n_samples, n_features].
            y: Labels of shape [n_samples].
        
        Returns:
            The object itself
        """
        self.sampler_list[0].fit(X, y)
        return self

    def select_samples(self, X: np.array,
                       sample_weight: np.array = None) -> np.array:
        """Selects the using uncertainty preselection and KMeans sampler.

        Args:
            X: Pool of unlabeled samples of shape (n_samples, n_features).
            sample_weight: Weight of the samples of shape (n_samples),
                optional.

        Returns:
            Indices of the selected samples of shape (batch_size).
        """
        selected = self.sampler_list[0].select_samples(X)
        kwargs = dict()
        if sample_weight is not None:
            kwargs['sample_weight'] = sample_weight[selected]
        new_selected = self.sampler_list[1].select_samples(
            X[selected], **kwargs)
        selected = selected[new_selected]
        
        return selected



import numpy as np

from .base import ScoredQuerySampler
from .typeutils import RandomStateType, check_random_state


class RandomSampler(ScoredQuerySampler):
    """Randomly select samples

    Args:
        batch_size : Number of samples to select.
        random_state : The seed of the pseudo random number generator to use
            when shuffling the data.  If int, random_state is the seed used by
            the random number generator; If RandomState instance, random_state
            is the random number generator; If None (defdault), the random
            number generator is the RandomState instance used by `np.random`.

    Attributes:
        random_state : The random state used by the sampler.
    """

    def __init__(self, batch_size: int, random_state: RandomStateType = None):
        super().__init__(batch_size=batch_size)
        self.random_state = random_state

    def fit(self, X: np.array = None, y: np.array = None) -> 'RandomSampler':
        """Sets the random state
        
        Args:
            X: Labeled samples of shape (n_samples, n_features).
            y: Labels of shape (n_samples).
        
        Returns:
            The object itself
        """
        self.random_state = check_random_state(self.random_state)
        return self

    def score_samples(self, X: np.array) -> np.array:
        return self.random_state.rand(X.shape[0])



# *- encoding: utf-8 -*-
"""
cardinal version, required package versions, and utilities for checking
"""
# Author: Alexandre Abraham

# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
# X.Y
# X.Y.Z # For bugfix releases
#
# Admissible pre-release markers:
# X.YaN # Alpha release
# X.YbN # Beta release
# X.YrcN # Release Candidate
# X.Y # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'
#
__version__ = '0.0.5'

# This is a tuple to preserve order, so that dependencies are checked
# in some meaningful order (more => less 'core').
DEPENDENCIES_METADATA = (
    ('numpy', {'min_version': '1.11'}),
    ('scipy', {'min_version': '0.19'}),
    ('scikit-learn', {
        'min_version': '0.19',
        'extra_options': ['sklearn', 'examples', 'submodular', 'doc']}),
    ('apricot-select', {
        'min_version': '0.5.0', 'extra_options': ['submodular', 'doc']}),
    ('matplotlib', {
        'min_version': '2.0', 'extra_options': ['examples', 'doc']}),
    ('sphinx-gallery', {
        'min_version': '0.5.0', 'extra_options': ['doc']}),
    ('sphinx', {
        'min_version': '2.2.2', 'extra_options': ['doc']}),
)

package_to_module = {
    'scikit-learn': 'sklearn',
    'apricot-select': 'apricot'
}


def check_modules(extra_option=None, import_module=None, strict=True):
    """Check that module is installed with a recent enough version

    Args:
        extra_option: If None, check based modules, otherwise checks
            modules for the specified option. None by default.
        import_module: If the check is made in a specific module, adds
            it to error messages. Empty by default.
        strict: If True (default), raises an error in case of problem.
            Otherwise returns a boolean indicating if the set up is ok.
    """
    from distutils.version import LooseVersion

    if import_module:
        import_module = '.' + import_module

    for package_name, metadata in DEPENDENCIES_METADATA:

        if not ((extra_option is None and 'extra_options' not in metadata)
                or (extra_option in metadata.get('extra_options', []))):
            continue

        min_version = metadata['min_version']
        try:
            module_name = package_to_module.get(package_name, package_name)
            module = __import__(module_name)
        except ImportError as exc:
            user_friendly_info = (
                'Module "{0}" could not be found. '
                'Please install it properly to use cardinal{1}.'.format(
                    package_name, import_module))
            exc.args += (user_friendly_info,)
            exc.msg += '. ' + user_friendly_info
            if strict:
                raise
            else:
                return False

        # Avoid choking on modules with no __version__ attribute
        module_version = getattr(module, '__version__', '0.0.0')

        version_too_old = (not LooseVersion(module_version) >=
                           LooseVersion(min_version))

        if version_too_old:
            message = (
                'A {package_name} version of at least {minimum_version} '
                'is required to use cardinal{import_module}. '
                '{module_version} was found. '
                'Please upgrade {package_name}').format(
                    package_name=package_name,
                    minimum_version=min_version,
                    module_version=module_version,
                    import_module=import_module)

            if strict:
                raise ImportError(message)
            else:
                return False
    if not strict:
        return True



from scipy.stats import entropy
import numpy as np

from .base import ScoredQuerySampler
from .typeutils import check_proba_estimator


def _get_probability_classes(
        classifier,
        X: np.ndarray) -> np.ndarray:
    """Returns classifier.predict_proba(X)

    Args:
        classifier: The classifier for which probabilities are to be queried.
        X: Samples to classify.

    Returns:
        The probability of each class for each sample.
    """
    if classifier == 'precomputed':
        return X
    check_proba_estimator(classifier)
    if classifier.__class__.__module__.split('.')[0] == 'keras':  # Keras models have no predict_proba
        classwise_uncertainty = classifier.predict(X)
    else:  # sklearn compatible model
        classwise_uncertainty = classifier.predict_proba(X)
    return classwise_uncertainty


def confidence_score(classifier, X: np.ndarray) -> np.ndarray:
    """Measure the confidence score of a model for a set of samples.

    Args:
        classifier: The classifier for which the labels are to be queried.
        X: The pool of samples to query from.

    Returns:
        The confidence score for each sample.
    """
    classwise_uncertainty = _get_probability_classes(classifier, X)
    uncertainty = 1 - np.max(classwise_uncertainty, axis=1)
    return uncertainty


def margin_score(classifier, X: np.ndarray) -> np.ndarray:
    """Compute the difference between the two top probability classes for each sample. 

    This strategy takes the probabilities of top two classes and uses their
    difference as a score for selection.

    Args:
        classifier: The classifier for which the labels are to be queried.
        X: The pool of samples to query from.

    Returns:
        The margin score for each sample.
    """
    classwise_uncertainty = _get_probability_classes(classifier, X)
    part = np.partition(classwise_uncertainty, -2, axis=1)
    margin = 1 - (part[:, -1] - part[:, -2])
    return margin


def entropy_score(classifier, X: np.ndarray) -> np.ndarray:
    """Entropy sampling query strategy, uses entropy of all probabilities as score.

    This strategy selects the samples with the highest entropy in their prediction
    probabilities.
    
    Args:
        classifier: The classifier for which the labels are to be queried.
        X: The pool of samples to query from.
        n_instances: Number of samples to be queried.

    Returns:
        The entropy score for each label
    """
    classwise_uncertainty = _get_probability_classes(classifier, X)
    entropies = np.transpose(entropy(np.transpose(classwise_uncertainty)))
    return entropies


class ConfidenceSampler(ScoredQuerySampler):
    """Selects samples with lowest prediction confidence.

    Lowest confidence sampling looks at the probability of the class predicted by
    the classifier and selects the samples where this probability is the lowest.

    Parameters:
        classifier: Classifier used to determine the prediction confidence.
            The object must comply with scikit-learn interface and expose a
            `predict_proba` method.
        batch_size: Number of samples to draw when predicting.
        assume_fitted: If true, classifier is not refit
        verbose: The verbosity level. Defaults to 0.
    
    Attributes:
        classifier_: The fitted classifier.
    """
    def __init__(self, classifier, batch_size: int,
                 strategy: str = 'top', assume_fitted: bool = False,
                 verbose: int = 0):
        super().__init__(batch_size, strategy=strategy)
        self.classifier_ = classifier
        self.assume_fitted = assume_fitted
        self.verbose = verbose
        if self.classifier_ == 'precomputed':
            self.assume_fitted = True
        else:
            check_proba_estimator(classifier)

    def fit(self, X: np.array, y: np.array) -> 'ConfidenceSampler':
        """Fit the estimator on labeled samples.

        Args:
            X: Labeled samples of shape (n_samples, n_features).
            y: Labels of shape (n_samples).
        
        Returns:
            The object itself
        """
        if not self.assume_fitted:
            self.classifier_.fit(X, y)
        return self

    def score_samples(self, X: np.array) -> np.array:
        """Selects the samples to annotate from unlabeled data.

        Args:
            X: shape (n_samples, n_features), Samples to evaluate.

        Returns:
            The score of each sample according to lowest confidence estimation.
        """
        return confidence_score(self.classifier_, X)


class MarginSampler(ScoredQuerySampler):
    """Selects samples with greatest confusion between the top two classes.

    Smallest margin sampling uses the difference of predicted probability between
    the top two classes to select the samples on which the model is hesitating
    the most, hence the lowest difference.

    Parameters:
        classifier: Classifier used to
            determine the prediction confidence. The object must
            comply with scikit-learn interface and expose a
            `predict_proba` method.
        batch_size: Number of samples to draw when predicting.
        assume_fitted: If true, classifier is not refit
        verbose: The verbosity level. Defaults to 0.
    
    Attributes:
        classifier_: The fitted classifier.
    """
    def __init__(self, classifier, batch_size: int,
                 strategy: str = 'top', assume_fitted: bool = False,
                 verbose: int = 0):
        super().__init__(batch_size, strategy=strategy)
        self.classifier_ = classifier
        self.assume_fitted = assume_fitted
        self.verbose = verbose
        if self.classifier_ == 'precomputed':
            self.assume_fitted = True
        else:
            check_proba_estimator(classifier)

    def fit(self, X: np.array, y: np.array) -> 'MarginSampler':
        """Fit the estimator on labeled samples.

        Args:
            X: Labeled samples of shape (n_samples, n_features).
            y: Labels of shape (n_samples).
        
        Returns:
            The object itself
        """
        if not self.assume_fitted:
            self.classifier_.fit(X, y)
        return self

    def score_samples(self, X: np.array) -> np.array:
        """Selects the samples to annotate from unlabeled data.

        Args:
            X: shape (n_samples, n_features), Samples to evaluate.

        Returns:
            The score of each sample according to lowest confidence estimation.
        """
        return margin_score(self.classifier_, X)


class EntropySampler(ScoredQuerySampler):
    """Selects samples with greatest entropy among all class probabilities.

    Greatest entropy sampling measures the uncertainty of the model over all
    classes through the entropy of the probabilites of all classes. Highest
    entropy samples are selected.

    Parameters:
        classifier: Classifier used to
            determine the prediction confidence. The object must
            comply with scikit-learn interface and expose a
            `predict_proba` method.
        batch_size: Number of samples to draw when predicting.
        assume_fitted: If true, classifier is not refit
        verbose: The verbosity level. Defaults to 0.
    
    Attributes:
        classifier_: The fitted classifier.
    """
    def __init__(self, classifier, batch_size: int,
                 strategy: str = 'top', assume_fitted: bool = False,
                 verbose: int = 0):
        super().__init__(batch_size, strategy=strategy)
        self.classifier_ = classifier
        self.assume_fitted = assume_fitted
        self.verbose = verbose
        if self.classifier_ == 'precomputed':
            self.assume_fitted = True
        else:
            check_proba_estimator(classifier)

    def fit(self, X: np.array, y: np.array) -> 'EntropySampler':
        """Fit the estimator on labeled samples.

        Args:
            X: Labeled samples of shape (n_samples, n_features).
            y: Labels of shape (n_samples).
        
        Returns:
            The object itself
        """
        if not self.assume_fitted:
            self.classifier_.fit(X, y)
        return self

    def score_samples(self, X: np.array) -> np.array:
        """Selects the samples to annotate from unlabeled data.

        Args:
            X: shape (n_samples, n_features), Samples to evaluate.

        Returns:
            The score of each sample according to lowest confidence estimation.
        """
        return entropy_score(self.classifier_, X)



from typing import Union, Callable
import numpy as np


RandomStateType = Union[np.random.RandomState, int, None]
MetricType = Union[Callable, str]


def check_random_state(seed: RandomStateType):
    """Turn seed into a np.random.RandomState instance
    
    Args:
    seed : If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.

    Note
    ----
    This was taken from scikit-learn and slightly modified
    """
    if isinstance(seed, int):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    return np.random.mtrand._rand


class NotEnoughSamplesWarning(UserWarning):
    """Custom warning used when a sampler is given less than batch_size samples
    """


def _has_method(obj, method):
    return hasattr(obj, method) and callable(getattr(obj, method))


def check_proba_estimator(obj):

    predict_name = 'predict_proba'

    # Special case, we allow keras models
    package = obj.__class__.__module__.split('.')[0]

    if package == 'keras':
        predict_name = 'predict'

    # If the object has the right method, it's ok
    has_fit = _has_method(obj, 'fit')
    has_predict_proba = _has_method(obj, predict_name)

    if has_fit and has_predict_proba:
        return

    origin = "object from package " + package

    if package == "__main__":
        origin = "user-defined object"

    raise TypeError('Provided {} does not have required methods fit and {}.'
                    ''.format(origin, predict_name))


import numpy as np

from .typeutils import check_random_state


def pad_with_random(array, size, min, max, random_state=None):
    n_missing = size - len(array)
    if n_missing <= 0:
        return array
    random_state = check_random_state(random_state)
    choices = np.arange(min, max)
    mask = np.ones(max - min, dtype=bool)
    mask[array - min] = False
    choices = choices[mask]
    padding = random_state.choice(choices, n_missing)
    return np.concatenate([array, padding])


class SampleSelector():

    def __init__(self, size):
        self.size = size
        self._mask = np.zeros((size,), dtype=np.bool)
        self._indices = np.arange(size)

    def add_to_selected(self, indices):
        self._mask[self._indices[~self._mask][indices]] = True

    @property
    def selected(self):
        v = self._mask.copy().view()
        v.setflags(write=False)
        return v

    @property
    def non_selected(self):
        v = (~self._mask).view()
        v.setflags(write=False)
        return v



from .version import check_modules

check_modules('submodular', 'submodularity')  # noqa

import numpy as np
from apricot import FacilityLocationSelection
from sklearn.metrics import pairwise_distances

from .base import BaseQuerySampler
from .typeutils import MetricType


class SubmodularSampler(BaseQuerySampler):
    """Select samples using a facility location selector
    
    Args:
        batch_size: Number of samples to select.
        metric: Metric to use for distance computation.
        n_jobs: Number of jobs to run in parallel. -1 means using all cores.
    """

    def __init__(self, batch_size: int, metric: MetricType = 'euclidean',
                 n_jobs: int = 1):
        super().__init__(batch_size)
        self.metric = metric
        self.n_jobs = n_jobs

    def fit(self, X: np.array, y: np.array = None) -> 'SubmodularSampler':
        """Does nothing.
        
        Args:
            X: Labeled samples of shape (n_samples, n_features).
            y: Labels of shape (n_samples).
        
        Returns:
            The object itself
        """
        return self

    def select_samples(self, X: np.array) -> np.array:
        """Select the best samples using submodular optimization.

        Args:
            X: Pool of unlabeled samples of shape (n_samples, n_features).

        Returns:
            Indices of the selected samples of shape (batch_size).
        """
        if self._not_enough_samples(X):
            return np.arange(X.shape[0])

        if self.metric != 'precomputed':
            distances = pairwise_distances(
                X, metric=self.metric, n_jobs=self.n_jobs)
            model = FacilityLocationSelection(
                self.batch_size, pairwise_func='precomputed').fit(distances)
        else:  
            model = FacilityLocationSelection(
                self.batch_size, pairwise_func='precomputed').fit(X)
        return model.ranking



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from cardinal.uncertainty import ConfidenceSampler, MarginSampler, EntropySampler


def test_all_uncertainty():

    proba = np.array([[0.10, 0.20, 0.30, 0.40],
                      [0.00, 0.05, 0.45, 0.50],
                      [0.15, 0.45, 0.20, 0.20]])

    # Confidence sampling choses the first sample
    sampler = ConfidenceSampler('precomputed', 1, assume_fitted=True)
    sampler.fit([], [])  # No training set, we consider it precomputed
    selected = sampler.select_samples(proba)
    assert_array_equal(selected, np.array([0]))

    # Margin sampling choses the second sample
    sampler = MarginSampler('precomputed', 1, assume_fitted=True)
    sampler.fit([], [])  # No training set, we consider it precomputed
    selected = sampler.select_samples(proba)
    assert_array_equal(selected, np.array([1]))

    # Entropy sampling choses the third sample
    sampler = EntropySampler('precomputed', 1, assume_fitted=True)
    sampler.fit([], [])  # No training set, we consider it precomputed
    selected = sampler.select_samples(proba)
    assert_array_equal(selected, np.array([2]))


class WithAllMethods:

    def fit(X, y=None):
        pass

    def predict_proba(X):
        pass


class MissingMethods:

    def fit(X, y=None):
        pass


def test_types():
    ConfidenceSampler(WithAllMethods(), 1)
    with pytest.raises(TypeError):
        ConfidenceSampler(MissingMethods(), 1)



import logging
from abc import ABC, abstractmethod

import numpy as np


class BaseMonitor(ABC):
    """A monitor is a metric and a set of utils to record it and monitor it.

    Args:
        batch_size: If specified, a warning will be issued if batch_size is not correct
        tolerance: 

    """

    def __init__(self, batch_size=None, tolerance=None):
        self.batch_size = batch_size
        self.tolerance = tolerance
        self.reset()

    def reset(self):
        self.n_samples = []
        self.values = []

    def _append_n_samples(self, n_samples):
        self.n_samples.append(n_samples)
        if not self.batch_size or len(self.n_samples) <= 1:
            return
        this_batch_size = self.n_samples[-1] - self.n_samples[-2]
        if this_batch_size != self.batch_size:
            logging.warn(
                'Batch size of iteration {} is {} which is different'
                'from the reference batch size {}'.format(
                    len(self.n_samples), this_batch_size, self.batch_size
                )
            )

    @abstractmethod
    def accumulate(self, n_samples, value):
        pass

    @abstractmethod
    def get(self):
        pass

    def is_stalled(self, n_iter=1):
        if len(self.values) < n_iter + 1:
            return False
        for prev_v, curr_v in zip(self.values[-n_iter - 1:-1], self.values[-n_iter]):
            if np.abs(curr_v - prev_v) > self.tolerance:
                return False
        return True
        

class ContradictionMonitor(BaseMonitor):
    """Stores the amount of contradictions along an experiment

    We call contradiction the difference between predictions of two successive
    models on an isolated test set.
    """

    """Stores contradiction for a new iteration.

    Args:
        n_samples : Number of training samples
        probas_test : Predictions of shape (n_samples, n_classes)
    """
    def accumulate(self, n_samples: int, probas_test: np.array):
        if self.last_probas_test is not None:
            self.values.append(
                np.abs(probas_test - self.last_probas_test).sum())
            self._append_n_samples(n_samples)
        self.last_probas_test = probas_test

    """Returns the recorded metrics
    """
    def get(self):
        return {
            "n_samples": self.n_samples,
            "contradictions": self.values
        }

    """Reset the metrics for a new experiment
    """
    def reset(self):
        super().reset()
        self.last_probas_test = None



# Dataiku Active Learning lib

# Cardinals, in the family Cardinalidae, are passerine birds found in
# North and South America. They are also known as cardinal-grosbeaks
# and cardinal-buntings.

from .version import __version__



#! /usr/bin/env python

import sys
import os
from collections import defaultdict

from setuptools import setup, find_packages


descr = """Active learning python package"""


def load_version():
    """Executes cardinal/version.py in a globals dictionary and return it.
    """
    globals_dict = {}
    with open(os.path.join('cardinal', 'version.py')) as fp:
        exec(fp.read(), globals_dict)
    return globals_dict


def is_installing():
    # Allow command-lines such as "python setup.py build install"
    install_commands = set(['install', 'develop'])
    return install_commands.intersection(set(sys.argv))


# Make sources available using relative paths from this file's directory.
os.chdir(os.path.dirname(os.path.abspath(__file__)))

_VERSION_GLOBALS = load_version()
DISTNAME = 'cardinal'
DESCRIPTION = 'Active learning python package'
with open('README.md') as fp:
    LONG_DESCRIPTION = fp.read()
MAINTAINER = 'Alexandre Abraham'
MAINTAINER_EMAIL = 'alexandre.abraham@dataiku.com'
URL = 'https://dataiku-research.github.io/cardinal'
LICENSE = 'Apache 2.0'
DOWNLOAD_URL = 'https://github.com/dataiku-research/cardinal'
VERSION = _VERSION_GLOBALS['__version__']


if __name__ == "__main__":
    if is_installing():
        module_check_fn = _VERSION_GLOBALS['check_modules']
        module_check_fn()

    install_requires = []
    extras_require = defaultdict(list)

    for mod, meta in _VERSION_GLOBALS['DEPENDENCIES_METADATA']:
        dep_str = '%s>=%s' % (mod, meta['min_version'])
        if 'extra_options' in meta:
            for extra_option in meta['extra_options']:
                extras_require[extra_option].append(dep_str)
            extras_require['all'].append(dep_str)
        else:
            install_requires.append(dep_str)
  
    setup(name=DISTNAME,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=VERSION,
          download_url=DOWNLOAD_URL,
          long_description=LONG_DESCRIPTION,
          long_description_content_type='text/markdown',
          zip_safe=False,  # the package can run out of an .egg file
          classifiers=[
              'Intended Audience :: Science/Research',
              'Intended Audience :: Developers',
              'Programming Language :: Python',
              'Topic :: Scientific/Engineering',
              'Operating System :: Microsoft :: Windows',
              'Operating System :: POSIX',
              'Operating System :: Unix',
              'Operating System :: MacOS',
              'Programming Language :: Python :: 3.5',
              'Programming Language :: Python :: 3.6',
              'Programming Language :: Python :: 3.7',
          ],
          packages=find_packages(),
          package_data={},
          python_requires='>=3.5',
          install_requires=install_requires,
          extras_require=extras_require)



from importlib import import_module
import dataset
import shelve
import copy
import numpy as np
import pickle
import os
import shutil
import itertools
from pathlib import Path
import json
import time
import pandas as pd
from abc import ABC, abstractmethod


class HashableDict(dict):
    def __key(self):
        return tuple((k, self[k]) for k in sorted(self))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()


class ValueStore(ABC):

    @abstractmethod
    def __init__(self, filename):
        pass

    @abstractmethod
    def _store(self, name, value, keys):
        pass

    @abstractmethod
    def _sync(self):
        pass

    @abstractmethod
    def get(self, name):
        pass

    @abstractmethod
    def close(self):
        pass


class ShelveStore(ValueStore):

    def __init__(self, filename, writeback=False):
        self.writeback = writeback
        import shelve
        if not filename.endswith('.db'):
            raise ValueError('File extension must be .db with shelve backend')
        self._conn = shelve.open(filename, writeback=writeback)

    def _store(self, name, value, **keys):
        keys = HashableDict(keys)
        if self.writeback:
            if not name in self._conn:
                self._conn[name] = dict()
            self._conn[name][keys] = value
        else: 
            d = dict()
            if name in self._conn:
                d = self._conn[name]
            d[keys] = value
            self._conn[name] = d

    def get(self, name):
        self._conn.sync()
        if name not in self._conn:
            return pd.DataFrame([])
        return pd.DataFrame.from_records([dict(value=v, **k) for k, v in self._conn[name].items()])

    def _sync(self):
        self._conn.sync()

    def close(self):
        self._conn.close()



class SqliteStore(ValueStore):

    def __init__(self, filename):
        import dataset
        self._conn = dataset.connect('sqlite:///' + filename)

    def _store(self, name, value, **keys):
        table = self._conn[name]
        table.upsert(dict(value=value, **keys), list(keys.keys()))

    def get(self, name):
        if name not in self._conn:
            return pd.DataFrame([])
        return pd.DataFrame(list(self._conn[name].all())).drop('id', axis=1)

    def _sync(self):
        self._conn.sync()

    def close(self):
        self._conn.close()


class ResumeCache:

    _clear_outdated_variables = True

    def __init__(self, cache_dir, value_store, keys={}):
        self.cache_dir = Path(cache_dir).joinpath(*[Path(k) / Path(str(v)) for k, v in keys.items()])
        self.value_store = value_store
        self.keys = keys
        self._current_iter = -1

    def __enter__(self):
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        return self

    def __exit__(self, type, value, traceback):
        ReplayCache.cache = None

    def persisted_value(self, name, init_value):
        """Value persisted in cache for resuming or replaying

        A persisted value is kept in memory and on disk for a short duration to allow
        for resuming (using ResumeCache), or forever to allow for replaying the
        experiment (ReplayCache).

        Args:
            name: The name of the persisted variable.
            init_value: The value attributed to the variable.

        Returns:
            A variable object allowing to access persisted data.
        """
        return Variable(name, init_value, cache=self.cache_dir, clear_outdated=self._clear_outdated_variables)

    def log_value(self, key, value, iteration='auto', **kwargs):
        if type(value).__module__ == np.__name__:
            value = value.item()
        log_keys = self.keys.copy()
        if iteration == 'auto' and self._current_iter >= 0:
            log_keys['iteration'] = self._current_iter
        elif iteration is not None:
            log_keys['iteration'] = iteration
        log_keys.update(kwargs)
        self.value_store._store(key, value, **log_keys)

    def iter(self, iterator, *variables):

        if self._current_iter == -2:
            raise ValueError('This experiment has already run')

        json_path = self.cache_dir / "progress.json"

        if json_path.exists():
            with open(json_path, "r") as json_file:
                data = json.load(json_file)
                self._current_iter = max([int(k) for k in data.keys()])

        for i, item in enumerate(iterator):
            if i <= self._current_iter:
                continue
            self._current_iter = i
            t0 = time.time()

            yield tuple([item] + [variable._get(i + offset) for variable, offset in variables])

            duration = time.time() - t0

            for variable, _ in variables:
                variable._persist(i)

            data = {}
            if json_path.exists():
                with open(json_path, "r") as json_file:
                    data = json.load(json_file)

            data[i] = duration

            with open(json_path, "w") as json_file:
                json.dump(data, json_file)

            for variable, _ in variables:
                variable._clear(i)
            
            self.value_store._sync()
        
        self._current_iter = -2


class ReplayCache(ResumeCache):

    _clear_outdated_variables = False

    def persisted_value(self, name, init_value):
        return Variable(name, init_value, cache=self.cache_dir, clear_outdated=False)

    def _is_variable(self, v):
        return (isinstance(v, tuple) and len(v) == 2 and isinstance(v[0], Variable))

    def compute_metric(self, name, func, *args, **kwargs):
        if not self._current_iter == -2:
            raise ValueError('Please run your experiment before computing metrics')

        json_path = self.cache_dir / "progress.json"
        with open(json_path, "r") as json_file:
            progress = json.load(json_file)
        max_iter = max([int(s) for s in progress])

        var_args = [i for i, a in enumerate(args) if self._is_variable(a)]
        var_kwargs = [k for k in kwargs if self._is_variable(kwargs[k])]

        for i in range(max_iter + 1):
            cargs = list(args)
            for j in var_args:
                cargs[j] = args[j][0]._get(args[j][1] + i)
                args[j][0]._clear(i)
            ckwargs = kwargs.copy()
            for k in var_kwargs:
                ckwargs[k] = kwargs[k][0]._get(kwargs[k][1] + i)
                ckwargs[k]._clear(i)
            v = func(*cargs, **ckwargs)
            self.log_value(name, v, iteration=i)


class Variable():

    def __init__(self, name, init_value, cache='./cache', clear_outdated=True):
        self.name = name
        self.init_value = init_value
        self.cache = Path(cache)
        self.clear_outdated = clear_outdated
        self._lifetime = 1
        self._values = dict()
    
    def previous(self, offset=-1):
        self._lifetime = max(-offset, self._lifetime)
        return self, offset

    def current(self):
        return self, 0

    def set(self, value):
        self._cached = value

    def _exists(self, iteration):
        return iteration < 0 or iteration in self._values or (self.cache / str(iteration) / self.name).exists()

    def _load(self, iteration):
        path = (self.cache / str(iteration) / self.name)

        # Check if numpy
        try:
            self._values[iteration] = np.load(path)
            return
        except ValueError:
            pass

        with open(path, 'rb') as filedesc:
            self._values[iteration] = pickle.load(filedesc)

    def _get(self, iteration):
        if iteration < 0:
            return self.init_value

        if not iteration in self._values:
            # Load from disk
            self._load(iteration)
        
        return self._values[iteration]

    def _persist(self, iteration):

        self._values[iteration] = self._cached
        
        path = (self.cache / str(iteration))
        path.mkdir(parents=True, exist_ok=True)
        path = path / self.name
        with open(path, 'wb') as f:
            if type(self._cached).__module__ == np.__name__:
                np.save(f, self._cached)
            else:
                pickle.dump(self._cached, f)

    def _clear(self, iteration):

        # Delete unnecessary objects
        to_delete = []
        for key in self._values:
            if key <= iteration - self._lifetime:
                to_delete.append(key)
        for key in to_delete:
            del self._values[key]
            if self.clear_outdated:
                path = (self.cache / str(key)) / self.name
                path.unlink()
